/*
 * deviced
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <malloc.h>
#include <stdlib.h>
#include <unistd.h>
#include <poll.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/eventfd.h>
#include <pthread.h>

#include <linux/usb/functionfs.h>
#include <systemd/sd-daemon.h>

#define FFS_PATH "/tmp/usb-host-test-ffs"

enum {
	EP0_IDX = 0,
	EP_IN_BULK_IDX,
	EP_OUT_BULK_IDX,
	EP_IN_INT_IDX,
	EP_OUT_INT_IDX,
	NUM_ENDPOINTS,
};

#define MAX(a, b) ((a) > (b) ? (a) : (b))

#define report_error(...) do {			\
		fprintf(stderr, __VA_ARGS__);	\
		putchar('\n');			\
	} while (0)

#define BUFSIZE 512
char int_buf[BUFSIZE];
char bulk_buf[BUFSIZE];

/* Close all ep files */
void cleanup_ffs(int *ep)
{
	int i;

	for (i = 0; i < NUM_ENDPOINTS; ++i)
		close(ep[i]);
}

/* Handle events generated by kernel and provided via ep0 */
int handle_ep0(int ep0, int *connected)
{
	struct usb_functionfs_event event;
	int ret;

	ret = read(ep0, &event, sizeof(event));
	if (!ret) {
		report_error("unable to read event from ep0");
		return -EIO;
	}

	switch (event.type) {
	case FUNCTIONFS_SETUP:
		/* stall for all setuo requests */
		if (event.u.setup.bRequestType & USB_DIR_IN) {
			if (write(ep0, NULL, 0) < 0)
				report_error("write error %d", errno);
		} else {
			if (read(ep0, NULL, 0) < 0)
				report_error("read error %d", errno);
		}
		break;

	case FUNCTIONFS_ENABLE:
		*connected = 1;
		break;

	case FUNCTIONFS_DISABLE:
		*connected = 0;
		break;

	default:
		break;
	}

	return 0;
}

struct tdata {
	int ep_in;
	int ep_out;
	char *buf;
};

void *loop_thread(void *data)
{
	struct tdata *d = data;
	int ret;

	while (1) {
		ret = read(d->ep_out, d->buf, BUFSIZE);
		if (ret < 0) {
			report_error("read error on fd %d: %m\n", d->ep_out);
			if (errno != ECONNRESET && errno != ESHUTDOWN)
				exit(ret);

			return NULL;
		}

		fprintf(stderr, "received %d bytes\n", ret);

		ret = write(d->ep_in, d->buf, ret);
		if (ret < 0) {
			report_error("write error on fd %d: %m\n", d->ep_in);
			if (errno != ECONNRESET && errno != ESHUTDOWN)
				exit(ret);

			return NULL;
		}

		fprintf(stderr, "sent %d bytes\n", ret);
	}

	return NULL;
}

/* main chat function */
void do_chat(int *ep)
{
	int connected = 0;
	int ret;
	pthread_t int_thread, bulk_thread;
	struct tdata int_data, bulk_data;

	int_data.ep_in = ep[EP_IN_INT_IDX];
	int_data.ep_out = ep[EP_OUT_INT_IDX];
	int_data.buf = int_buf;
	bulk_data.ep_in = ep[EP_IN_BULK_IDX];
	bulk_data.ep_out = ep[EP_OUT_BULK_IDX];
	bulk_data.buf = bulk_buf;

	while (1) {
		while (!connected) {
			ret = handle_ep0(ep[0], &connected);
			if (ret < 0)
				return;
		}

		ret = pthread_create(&bulk_thread, NULL, loop_thread, &bulk_data);
		if (ret < 0)
			return;

		ret = pthread_create(&int_thread, NULL, loop_thread, &int_data);
		if (ret < 0) {
			pthread_cancel(bulk_thread);
			return;
		}

		pthread_join(bulk_thread, NULL);
		pthread_join(int_thread, NULL);

		connected = 0;
	}
}

int main(int argc, char **argv)
{
	int ep[NUM_ENDPOINTS];
	int i = 0;

	if (sd_listen_fds(0) < NUM_ENDPOINTS) {
		fprintf(stderr, "Expected endpoints not retrieved\n");
		return -1;
	}

	for (i = 0; i < NUM_ENDPOINTS; ++i)
		ep[i] = SD_LISTEN_FDS_START + i;

	do_chat(ep);
	cleanup_ffs(ep);

	return 0;
}
